#include <nds.h>
#include <nds/arm9/console.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <user_debugger.h>
#include <fat.h>

#include <dswifi9.h>

#include "sys.h"

#include "quake_ipc.h"
#include "fixed_point.h"

//#define USE_WIFI
//#define USE_DEBUGGER
#define USE_3D
//#define USE_MP3
//#define USE_QUAKE_CONSOLE
#define FAKE_QUAKE_CONSOLE
#define USE_REAL_TIMEBASE
#define DO_FP_TESTS

//#define TEXTURES_HAVE_NAMES
#define TEXTURE_CACHE_IN_VRAM

//#define IPC_IN_HBLANK
//#define IPC_IN_VBLANK
#define IPC_IN_TIMER

//#define USE_EXTRA_RAM
#define ARM7_USES_EXRAM


#ifdef USE_DEBUGGER
#define DEBUG_MSG(x) debug_msg(x)
#else
#define DEBUG_MSG(x)
#endif

extern void quake_main (int argc, char **argv);

extern "C" void debug_print_stub(char* string)
{
	printf(string);
}

volatile unsigned int decoder_stopped = 1;
volatile unsigned int in_file_stuff = 0;

#ifdef DO_FP_TESTS
#define FABS(x) ((x) < 0 ? (-(x)) : ((x)))
void test_number(float f, bool fail_expected) 
{ 
	float f1 = fixed_to_float(float_to_fixed(f)); 
	float f2 = fixed_to_float(float_to_fixed_fpu(f)); 
	bool fail = false; 

	if (FABS(f1 - f) > 1.f / (1 << FP_SCALE_SHIFT)) fail = true; 
	if (f1 != f2) fail = true; 

	fail = (fail == !fail_expected);
	if (fail)
	{ 
		printf("%s %f -> %f (%f) / %f (%f)\n", false == fail ? "SUCCESS" : "FAIL", f, f1, FABS(f1 - f), f2, FABS(f2 - f));
		Sys_Error("failed fp tests\n");
	} 
}
#undef FABS

void do_fp_tests(void)
{
	test_number(0.01f, false); 
   test_number(0.1f, false); 
   test_number(128.1f, false); 
   test_number(99999.1f, false); 
   /* biggest representable value */ 
   test_number(float((1 << 18) - 1), false); 
   /* this is expected to fail, since it can't be represented in the fixed point format */ 
   test_number(float((1 << 18)), true); 

   test_number(1e-20, false); 


   test_number(-0.01f, false); 
   test_number(-0.1f, false); 
   test_number(-128.1f, false); 
   test_number(-99999.1f, false); 
   test_number(-float((1 << 18) - 1), false); 
   /* this is expected to work, since it can be represented in the fixed point format (there's one more encoding in negative values)*/ 
   test_number(-float((1 << 18)), false); 

   /* this should fail */ 
   test_number(-float((1 << 18) + 1), true); 
   
   printf("passed fp tests!\n");
//   while(1);
}
#endif

void ds_exram_unlock(void);
void ds_exram_lock(void);
void ds_try_exram_unlock(void);
void ds_try_exram_lock(void);
void *ds_exram_base(void);
unsigned int ds_exram_size(void);

void file_lock(void)
{
	int lock_count = 0;
	while (in_file_stuff == 1)
	{
		if (lock_count > 1000000)
		{
			printf("trying to lock file access\n");
			lock_count = 0;
		}
		lock_count++;
	}
	
	in_file_stuff = 1;
	
	ds_try_exram_lock();
//	printf("file access locked\n");
}

void file_unlock(void)
{
//	printf("file access unlocked\n");
	if (in_file_stuff == 0)
		Sys_Error("trying to unlock memory - already unlocked!\n");
	in_file_stuff = 0;
	ds_try_exram_unlock();
}

void ipc_block_ready_9to7(void)
{
	while (quake_ipc_9to7->message == 0xffffffff);
}

bool ipc_test_ready_9to7(void)
{
	return (quake_ipc_9to7->message == 0xffffffff);
}

void ipc_set_ready_9to7(void)
{
	quake_ipc_9to7->message = 0xffffffff;
}

void ipc_block_ready_7to9(void)
{
	while (quake_ipc_7to9->message == 0xffffffff);
}

unsigned char *S_LoadSound(char *name);
extern int com_filesize;

unsigned int time_to_sleep = 0;

bool arm7_ping = false;
void handle_ipc(void)
{
	if (in_file_stuff)
		return;
	
	if (quake_ipc_7to9->message == 0xffffffff)
	{
		switch (quake_ipc_7to9->message_type)
		{
			case kPrintMessage:
			{
				printf("%s", quake_ipc_7to9_buf);
				break;
			}
			case kPrintNumber:
			{
				printf("ARM7: %08x", *(unsigned int *)quake_ipc_7to9_buf);
				break;
			}
			case kPrintFloat:
			{
				printf("ARM7: %d", (int)*(volatile float *)quake_ipc_7to9_buf);
				break;
			}
			case kMalloc:
			{
				printf("ARM7 malloc req %d bytes\n", *(unsigned int *)quake_ipc_7to9_buf);
				unsigned int addr = (unsigned int)malloc(*(unsigned int *)quake_ipc_7to9_buf);
				
				if (addr == 0)
					printf("ARM7 MALLOC FAILED!\n");
				
				addr |= 0x400000;
				
				quake_ipc_7to9->message_type = kMallocResponse;
				*(unsigned int *)quake_ipc_7to9_buf = addr;
				break;
			}
			case kFree:
			{
				printf("ARM7 free req on %08x\n", *(unsigned int *)quake_ipc_7to9_buf);
				void *ptr = (void *)*(unsigned int *)quake_ipc_7to9_buf;
				free(ptr);
				break;
			}
			case kFOpen:
			{
				file_lock();
				
				printf("ARM7 fopen: %s\n", quake_ipc_7to9_buf);
				FILE *fp = fopen((char *)quake_ipc_7to9_buf, "rb");
				
				if (fp)
					printf("...ok\n");
				else
					printf("...failed\n");
				
				printf("9: fp is %08x\n", (unsigned int)fp);
				
				*(unsigned int *)quake_ipc_9to7_buf = (unsigned int)fp;
				
				file_unlock();
				
				break;
			}
			case kFClose:
			{
				file_lock();
				
				printf("ARM7 fclose: %08x\n", *(unsigned int *)quake_ipc_7to9_buf);
				unsigned int fp = *(unsigned int *)quake_ipc_7to9_buf;
				fclose((FILE *)fp);
				
				file_unlock();
				
				break;
			}
			case kFRead:
			{
				file_lock();
				
				unsigned int buf = ((unsigned int *)quake_ipc_7to9_buf)[0];
				unsigned int size = ((unsigned int *)quake_ipc_7to9_buf)[1];
				unsigned int num = ((unsigned int *)quake_ipc_7to9_buf)[2];
				unsigned int fp = ((unsigned int *)quake_ipc_7to9_buf)[3];
				
					printf("ARM7 fread(%08x %d %d %08x)\n",
						buf, size, num, fp);
				
				unsigned int result = fread((unsigned char *)buf, size, num, (FILE *)fp);
				*(unsigned int *)quake_ipc_9to7_buf = result;
				
				file_unlock();

				break;
			}
			case kStopMP3:
			{
				printf("decoder has stopped\n");
				decoder_stopped = 1;
				break;
			}
			case kS_LoadSound:
			{
//				printf("ARM9: S_LoadSound, %s\n", (char *)quake_ipc_7to9_buf);
				unsigned char *data = S_LoadSound((char *)quake_ipc_7to9_buf);
//				unsigned char *data = S_LoadSound("weapons/lhit.wav");
				
//				FILE *fp = fopen("fat0:/out.wav", "wb");
//				if (fp == NULL)
//					Sys_Error("failed to open file\n");
//				fwrite(data, 1, com_filesize, fp);
//				fclose(fp);
//				printf("done\n");
//				while(1);
				
				((unsigned int *)quake_ipc_7to9_buf)[0] = (unsigned int)data;	//yeah, I know
				((unsigned int *)quake_ipc_7to9_buf)[1] = com_filesize;
				
//				printf("ARM9: sound data is in %08x, size %d\n", (unsigned int)data, com_filesize);
//				printf("ARM9: read size as %d bytes\n", ((unsigned int *)data)[1] + 8);
				
				quake_ipc_7to9->message_type = kS_LoadSoundResponse;
				
				while (quake_ipc_7to9->message_type == kS_LoadSoundResponse);
				break;
			}
			case kHalt:
			{
				printf("ARM7 IS HALTING!\n");
				while(1);
				break;
			}
			case kLidHasClosed:
			{
				printf("lid has closed\n");
				time_to_sleep = 1;
				break;
			}
			case kPing:
			{
				arm7_ping = !arm7_ping;
				unsigned short *text_map = (unsigned short *)(((15)*0x800)+0x6200000);
		
				if (arm7_ping)
					text_map[32 * 23] = 0xf058;
				else
					text_map[32 * 23] = 0xf02b;
					
				break;
			}	
		}
		
		quake_ipc_7to9->message = 0;
	}
}

/*
===============================================================================

FILE IO

===============================================================================
*/

#define MAX_HANDLES             10
FILE    *sys_handles[MAX_HANDLES];

int             findhandle (void)
{
	int             i;
	
	for (i=1 ; i<MAX_HANDLES ; i++)
		if (!sys_handles[i])
			return i;
	Sys_Error ("out of handles");
	return -1;
}


/*
================
filelength
================
*/
int filelength (FILE *f)
{
	file_lock();
	
	int             pos;
	int             end;

	pos = ftell (f);
	fseek (f, 0, SEEK_END);
	end = ftell (f);
	fseek (f, pos, SEEK_SET);

	file_unlock();
	
	return end;
}

int Sys_FileOpenRead (char *path, int *hndl)
{
//	printf("fopen %s\n", path);
	FILE    *f;
	int             i;
	
	i = findhandle ();

	file_lock();
	f = fopen(path, "rb");
	file_unlock();
	
	if (!f)
	{
		*hndl = -1;
//		printf("failed to open file %s\n", path);
		return -1;
	}
	sys_handles[i] = f;
	*hndl = i;
	
	int l = filelength(f);
	
	return l;
}

int Sys_FileOpenWrite (char *path)
{
	file_lock();
	
	FILE    *f;
	int             i;
	
	i = findhandle ();

	f = fopen(path, "wb");
	if (!f)
		Sys_Error ("Error opening %s", path);
	sys_handles[i] = f;
	
	file_unlock();
	
	return i;
}

void Sys_FileClose (int handle)
{
	file_lock();
	
	fclose (sys_handles[handle]);
	sys_handles[handle] = 0;
	
	file_unlock();
}

void Sys_FileSeek (int handle, int position)
{
	file_lock();
	fseek (sys_handles[handle], position, SEEK_SET);
//	printf("soke to %d (%d)\n", position, handle);
	file_unlock();
}

void Sys_FileSeekMore (int handle, int position, int whence)
{
	file_lock();
	fseek (sys_handles[handle], position, whence);
//	printf("more to %d (%d)\n", position, handle);
	file_unlock();
}

int Sys_FileRead (int handle, void *dest_r, int count)
{
	unsigned char *dest = (unsigned char *)dest_r;
	file_lock();
	int length_read = fread (dest, 1, count, sys_handles[handle]);
	if (length_read != count)
		Sys_Printf("read was too short: wanted %d got %d\n", count, length_read);
	
//	for (int count = 0; count < 4; count++)
//		printf("%02x%02x%02x%02x ", dest[count * 4 + 0], dest[count * 4 + 1], dest[count * 4 + 2], dest[count * 4 + 3]);
//	printf("\nread %d\n", length_read);
		
	file_unlock();
	return length_read;
}

//int mmap_file_handle = -1;
//unsigned mmap_file_length;
//
//void* mmap_openfile(char *filename)
//{
//	mmap_file_length = Sys_FileOpenRead(filename, &mmap_file_handle);
//	
//	if (mmap_file_length < 0)
//	{
//		Sys_Printf("failed to load in mmapped file, %s\n", filename);
//		return NULL;
//	}
//	else
//	{
//		Sys_Printf("file %s successfully mmapped\n", filename);
//		return (void *)0x10000000;
//	}
//}
//
//void mmap_closefile(void)
//{
//	 mmap_file_length = 0;
//	 Sys_FileClose(mmap_file_handle);
//}
//
unsigned int mmap_handle(unsigned ea, int direction, int size)
{
}
//{
//	unsigned int data;
////	ea = ea - 0x10000000;
//	
//	if (size == 0)
//		size = 4;
//	
//	Sys_Printf("handle %08x, %d, %d\n", ea, direction, size);
//	
//	if (direction == 1)			//load
//	{
////		Sys_FileSeek(mmap_file_handle, ea);
////		Sys_FileRead(mmap_file_handle, &data, size);
//		data = ea;
//		
////		Sys_Printf("data is %08x\n", data);
//	}
//	else if (direction == 0)	//store
//	{
//	}
//	else
//		Sys_Error("\nmemory direction is %d!\n", direction);
//	
////	while(1);
//	return data;
//}

int Sys_FileWrite (int handle, void *data, int count) __attribute__ ((no_instrument_function));
int Sys_FileWrite (int handle, void *data, int count)
{
	file_lock();
	int res = fwrite (data, 1, count, sys_handles[handle]);
	file_unlock();
	return res;
}

int Sys_FileTime (char *path)
{
	file_lock();
	FILE    *f;
	
	f = fopen(path, "r");
	if (f)
	{
		fclose(f);
		file_unlock();
		return 1;
	}
	
	file_unlock();
	
	return -1;
}

unsigned short *frame_buffer_base;
unsigned short *console_base;
int frame_buffer_width;
int frame_buffer_height;

int blanks = 0;

int arm7_initialised = 0;

void init_arm7(void)
{
	printf("Init ARM7...");
	quake_ipc_9to7_buf = (unsigned char *)malloc(100);
	quake_ipc_7to9_buf = (unsigned char *)malloc(100);
	
	if ((quake_ipc_9to7_buf == NULL) || (quake_ipc_7to9_buf == NULL))
	{
		Sys_Error("couldn\'t alloc ARM9<->ARM7 IPC space\n");
		while(1);
	}
	
	quake_ipc_9to7_buf = (unsigned char *)((unsigned int)quake_ipc_9to7_buf | 0x400000);
	quake_ipc_7to9_buf = (unsigned char *)((unsigned int)quake_ipc_7to9_buf | 0x400000);
	
//	printf("9: 9->7 %08x 7->9 %08x\n", quake_ipc_9to7_buf, quake_ipc_7to9_buf);
	
	//send the init message
	quake_ipc_9to7->message_type = kInit;
	ipc_set_ready_9to7();
	
	//the ARM7 will now change the message to say it's ready
	ipc_block_ready_9to7();
	
//	printf("9: 7 has begun init\n");
	
	quake_ipc_9to7->message = (unsigned int)quake_ipc_9to7_buf;
	
	//it'll change this when it's read it
	while (quake_ipc_9to7->message != 0xffffffff);
	
//	printf("9: 7 has read 9to7\n");
	
	quake_ipc_9to7->message = (unsigned int)quake_ipc_7to9_buf;
	
	//it'll change this when it's read it
	while (quake_ipc_9to7->message == (unsigned int)quake_ipc_7to9_buf);
	
//	printf("9: 7 has read 7to9\n");
	
	quake_ipc_9to7->message = 0;
	
	arm7_initialised = 1;
}

void stop_mp3(void)
{
#ifdef USE_MP3
	printf("stopping track\n");
	ipc_block_ready_9to7();
	
	quake_ipc_9to7->message_type = kStopMP3;
	ipc_set_ready_9to7();
	
	ipc_block_ready_9to7();
	
	while (decoder_stopped == 0);
#endif
}

void play_mp3(unsigned char track)
{
#ifdef USE_MP3
	if (decoder_stopped == 0)
		stop_mp3();
	
	decoder_stopped = 0;
	
	ipc_block_ready_9to7();
	printf("playing track %d\n", track);

	quake_ipc_9to7->message_type = kPlayMP3;
	sprintf((char *)quake_ipc_9to7_buf, "/quake soundtrack/0%d-AudioTrack 0%d.mp3", track, track);
	ipc_set_ready_9to7();
	
	ipc_block_ready_9to7();
#else
	decoder_stopped = 1;
#endif
}

void load_textures(void);

void SCR_CheckDrawCenterString (void);
void SCR_DrawConsole (void);
void Draw_Crosshair(void);

int update_count = 0;

unsigned short *new_palette = NULL;
void ds_loadpalette(unsigned short *palette);
extern int in_menu;


void vblank_handler(void)
{
#if (defined USE_QUAKE_CONSOLE) || (defined FAKE_QUAKE_CONSOLE)
	update_count++;
	if (update_count > 10)
	{
//		extern unsigned short *console_base;
//		int count;
//		for (count = 0; count < 128 * 128; count++)
//			console_base[count] = 0;
//		
//		extern int extra;
//		if (extra == 0)
//			extra = 1;
//		else
//			extra = 0;
		
		if (in_menu == 0)
		{
//			SCR_CheckDrawCenterString ();
//			SCR_DrawConsole ();
//			Draw_Crosshair();
//			printf("\n");
			
//			first_in_menu = 1;
//			
//			if (first_out_menu)
//			{
//				first_out_menu = 0;
////				printf("\x1b[2J");
//			}
		}
		else
		{
//			if (first_in_menu == 1)
//			{
//				first_in_menu = 0;
////				printf("\x1b[2J");
//			}
//			first_out_menu = 1;
		}
		
		update_count = 0;
	}
#endif

	if (new_palette != NULL)
	{
		ds_loadpalette(new_palette);
		new_palette = NULL;
	}
	
#ifdef IPC_IN_VBLANK
	handle_ipc();
#endif
	
	if (blanks & 0x1)
		*(u16*)SCREEN_BASE_BLOCK_SUB(15) = 0xf058;
	else
		*(u16*)SCREEN_BASE_BLOCK_SUB(15) = 0xf02b;
	
//	load_textures();
	
	blanks++;
}

extern unsigned int super_framecount;
unsigned int hblanks = 0;

void hblank_handler(void)
{
//	unsigned short vcount = *((volatile unsigned short *)0x4000006);
	hblanks++;
//	if (vcount > 261)
//		hblanks = 0;
#ifdef IPC_IN_HBLANK
	handle_ipc();
#endif
}

double hblank_ms = 1.0 / 15.7343;
unsigned int initial_hblank = -1;
double Sys_FloatTime (void)
{
#ifdef USE_REAL_TIMEBASE
	if (initial_hblank == -1)
		initial_hblank = hblanks;
	return (double)(hblanks - initial_hblank) / (262 * 60);
#else
	static double t;
	
	t += 0.1;
	
	return t;
#endif
}

void ds_dcacheflush(void)
{
	DC_FlushAll();
}

float z_tri = 0.0f;
float z_quad = 0.0f;

int flush_val = 2;

float left_move = 0;
float up_move = 0;
float fov = 50;
float x_scale = 160;
float y_scale = 160;
float z_scale = 80;
float z_move = 0;
int palette_address = 0;

void ds_ortho(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-90, 130, -100, 60, 1.0f, 0x7fff);
	glMatrixMode(GL_TEXTURE);
	glIdentity();
	glMatrixMode(GL_MODELVIEW);
}

void ds_perspective(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fov, 256.0 / 192.0, 4.0f, 0x7fff);
	glMatrixMode(GL_TEXTURE);
	glIdentity();
	glScalef(4, 4, 4);
	glMatrixMode(GL_MODELVIEW);
}

bool use_wireframe = false;
void ds_setwireframe(bool wire)
{
	use_wireframe = wire;
}

/* from a thread in gbadev */

#define POLY_FOG  (1<<15) 

#define GL_FOG_ALPHA    (1<<6) 
#define GL_FOG          (1<<7) 
#define GL_FOG_SHIFT(n)   ((n)<<8) 

#define GL_RDLINE_UNDERFLOW   (1<<12) 
#define GL_VERTEX_OVERFLOW    (1<<13) 

#define GFX_FOG_COLOR   (*(vuint32*) 0x04000358) 
#define GFX_FOG_OFFSET  (*(vuint32*) 0x0400035C) 
#define GFX_FOG_TABLE   ((vuint8*)  0x04000360)

void glFogColor(uint8 red, uint8 green, uint8 blue, uint8 alpha) { GFX_FOG_COLOR = (red) | ((green)<<5) | ((blue)<<10) | ((alpha)<<16); } 
void glFogDepth(uint16 depth) { GFX_FOG_OFFSET = depth; }

/*  */

int base_polyid = 0;
void ds_polyfmt(int poly_id, int trans, int depth)
{
	if (use_wireframe)
	{
		int p = (base_polyid++) & 0x1f;
		glPolyFmt(POLY_ALPHA(trans) | POLY_FOG | POLY_CULL_NONE | POLY_ID(p) | (depth << 14));
	}
	else
		glPolyFmt(POLY_ALPHA(trans) | POLY_FOG | POLY_CULL_NONE | POLY_ID(poly_id) | (depth << 14));
}

int use_fog = 0;
int fog_depth = 0x200;
int fog_shift = 2;

void ds_set_fog(int shift, int depth)
{
	fog_shift = shift;
	fog_depth = depth;
}

void ds_enableAA(void)
{
	glEnable(GL_ANTIALIAS);
}

uint32 vram_banks;
void ds_unlock_vram(void)
{
	vram_banks = vramSetMainBanks(VRAM_A_LCD,VRAM_B_LCD,VRAM_C_LCD,VRAM_D_LCD);
}

void ds_lock_vram(void)
{
	vramRestoreMainBanks(vram_banks);
}

void ds_reset_render(void)
{
	glReset();
}

unsigned char fog_red = 0xff;
unsigned char fog_green = 0xde;
unsigned char fog_blue = 0xad;

void ds_begin_render(void)
{
#ifdef USE_3D
//	glReset();
	glMatrixMode(GL_PROJECTION);
	
	ds_perspective();
	
	glMaterialf(GL_AMBIENT, RGB15(0,0,0));
	glMaterialf(GL_DIFFUSE, RGB15(0,0,0));
	glMaterialf(GL_SPECULAR, BIT(15) | RGB15(0,0,0));
	glMaterialf(GL_EMISSION, RGB15(0,0,0));

	//ds uses a table for shinyness..this generates a half-ass one
	glMaterialShinyness();
	
	glFogColor(fog_red >> 2, fog_green >> 2, fog_blue >> 2, 31);
//	glFogDepth(0x200); 
//	glFogDepth(0x60); 
	int i; 
	for (i = 0; i < 32; i++) { 
		GFX_FOG_TABLE[i] = i << 2; 
	}
	
	ds_polyfmt(0, 31, 0);
	glEnable(GL_BLEND);
	glMatrixMode(GL_TEXTURE);
	glIdentity();
//	glScalef(4, 4, 4);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	
	if (use_fog)
	{
		glFogDepth(fog_depth);
		glEnable((1 << 7) | (GL_FOG_SHIFT(fog_shift)));
	}
	
	glTranslatef(left_move, up_move, z_move);
	glScalef(x_scale, y_scale, z_scale);
#endif
}

void unload_textures_2(void);

#define HBLANK_START_LOAD 150
#define HBLANK_START_LATE_LOAD 200
#define HBLANK_END_LOAD 205

void ds_end_render(void)
{
#ifdef USE_3D
	glPopMatrix(1);
	
	while (*((volatile unsigned short *)0x4000006) < HBLANK_START_LOAD);
	
	load_textures();

//	ds_unlock_vram();
//	
//	while (*((volatile unsigned short *)0x4000006) < 214);
//	ds_lock_vram();
	
	//glFlush();
	GFX_FLUSH = flush_val;
	
//	unload_textures_2();
#endif
}

void ds_pushmatrix(void)
{
#ifdef USE_3D
	glPushMatrix();
#endif
}

void ds_popmatrix(void)
{
#ifdef USE_3D
	glPopMatrix(1);
#endif
}

void ds_translatef(float x, float y, float z)
{
#ifdef USE_3D
	glTranslatef(x, y, z);
#endif
}

void ds_scalef(float x, float y, float z)
{
#ifdef USE_3D
	glScalef(x, y, z);
#endif
}

void ds_rotateX(float angle)
{
#ifdef USE_3D
	glRotateX(angle);
#endif
}

void ds_rotateY(float angle)
{
#ifdef USE_3D
	glRotateY(angle);
#endif
}

void ds_rotateZ(float angle)
{
#ifdef USE_3D
	glRotateZ(angle);
#endif
}

void ds_usematrix3x3(float *up, float *right, float *forward)
{
#ifdef USE_3D
	int upf32[3], rightf32[3], forwardf32[3];
	int count;
	
	for (count = 0; count < 3; count++)
	{
		upf32[count] = floattof32(up[count]);
		rightf32[count] = floattof32(right[count]);
		forwardf32[count] = floattof32(forward[count]);
	}
	
//	MATRIX_MULT3x3 = upf32[0];
//	MATRIX_MULT3x3 = upf32[1];
//	MATRIX_MULT3x3 = upf32[2];
//	
//	MATRIX_MULT3x3 = rightf32[0];
//	MATRIX_MULT3x3 = rightf32[1];
//	MATRIX_MULT3x3 = rightf32[2];
//	
//	MATRIX_MULT3x3 = forwardf32[0];
//	MATRIX_MULT3x3 = forwardf32[1];
//	MATRIX_MULT3x3 = forwardf32[2];
	
	MATRIX_MULT3x3 = upf32[0];
	MATRIX_MULT3x3 = rightf32[0];
	MATRIX_MULT3x3 = forwardf32[0];
	
	MATRIX_MULT3x3 = upf32[1];
	MATRIX_MULT3x3 = rightf32[1];
	MATRIX_MULT3x3 = forwardf32[1];
	
	MATRIX_MULT3x3 = upf32[2];
	MATRIX_MULT3x3 = rightf32[2];
	MATRIX_MULT3x3 = forwardf32[2];
#endif
}

void ds_begin_triangle(void)
{
#ifdef USE_3D
	glBegin(GL_TRIANGLE);
#endif
}

void ds_begin_quad(void)
{
#ifdef USE_3D
	glBegin(GL_QUAD);
#endif
}

void ds_begin_triangle_strip(void)
{
#ifdef USE_3D
	glBegin(GL_TRIANGLE_STRIP);
#endif
}

void ds_end(void)
{
#ifdef USE_3D
	glEnd();
#endif
}

void ds_vertex3f(float x, float y, float z)
{
#ifdef USE_3D
	glVertex3f(x, y, z);
#endif
}

void ds_vertex3v16(short int x, short int y, short int z)
{
#ifdef USE_3D
	glVertex3v16(x, y, z);
#endif
}

void ds_color3f(float red, float green, float blue)
{
#ifdef USE_3D
	glColor3f(red, green, blue);
#endif
}

void ds_color3b(unsigned char red, unsigned char green, unsigned char blue)
{
#ifdef USE_3D
	glColor3b(red, green, blue);
#endif
}

void ds_texcoord2t16(short u, short v)
{
#ifdef USE_3D
	glTexCoord2t16(u, v);
#endif
}

void ds_texcoord2f(float u, float v)
{
#ifdef USE_3D
	glTexCoord2f(u, v);
#endif
}

void BindNoTexture(void);
void ResetMenuCache(void);
void Sbar_Init(void);
void SCR_Init(void);

extern int reloading_gui_textures;
int vram_used = 0;
void vram_init(void);
void unload_textures(void);
void ds_resettextures(void)
{
//#ifdef USE_3D
//	printf("Resetting textures\n");
//	glResetTextures();
//	vram_init();
//	vram_used = 0;
//	
//	BindNoTexture();
//	ResetMenuCache();
//	
//	reloading_gui_textures = 1;
//	Sbar_Init();
//	SCR_Init();
//	reloading_gui_textures = 0;
//	
//	ds_reset_render();
//	ds_begin_render();
//	ds_end_render();
//	ds_reset_render();
//	ds_begin_render();
//	ds_end_render();
//#endif

//#ifdef USE_3D
//	printf("Resetting textures\n");
//
	unload_textures();
//	
//	BindNoTexture();
//	ResetMenuCache();
//	
//	reloading_gui_textures = 1;
//	Sbar_Init();
//	SCR_Init();
//	reloading_gui_textures = 0;
//	
//	ds_reset_render();
//	ds_begin_render();
//	ds_end_render();
//	ds_reset_render();
//	ds_begin_render();
//	ds_end_render();
//#endif
}

int ds_gettexwidth(void)
{
#ifdef USE_3D
	int i;
	glGetInt(GL_GET_TEXTURE_WIDTH, &i);
	
	return i;
#else
	return 0;
#endif
}

int ds_gettexheight(void)
{
#ifdef USE_3D
	int i;
	glGetInt(GL_GET_TEXTURE_HEIGHT, &i);
	
	return i;
#else
	return 0;
#endif
}

extern unsigned int *textures;
extern unsigned int activeTexture;
void ds_bindtexture(int id)
{
#ifdef USE_3D
	glBindTexture(0, id);
#endif
}

int ds_gentexture(void)
{
#ifdef USE_3D
	int temp_texture;
	glGenTextures(1, &temp_texture);
	return temp_texture;
#endif
}

extern "C" int getNextPaletteSlot(unsigned short count, unsigned char format);

void ds_schedule_loadpalette(unsigned short *palette)
{
	new_palette = palette;
}

void ds_loadpalette(unsigned short *palette)
{
#ifdef USE_3D
	memcpy(BG_PALETTE_SUB, palette, 256 * 2);
	
	if (palette_address == 0)
	{
//		int addr = gluTexLoadPal(palette, 256, GL_RGB256);
		int addr = getNextPaletteSlot(256, GL_RGB256);
		  
		if( addr>=0 )
		{
			vramSetBankF(VRAM_F_LCD);
	 		swiCopy( palette, &VRAM_F[addr>>1] , 256 / 2 | COPY_MODE_WORD);
		 	vramSetBankF(VRAM_F_TEX_PALETTE);
		}
		
		if (addr == -1)
		{
			printf("couldn\'t load palette!");
			*(int *)0 = 0;
		}
		palette_address = addr;
	}
	else
//		glTexLoadPal(palette, 256, palette_address);
	{
		vramSetBankF(VRAM_F_LCD);
 		swiCopy( palette, &VRAM_F[palette_address>>1] , 256 / 2 | COPY_MODE_WORD);
	 	vramSetBankF(VRAM_F_TEX_PALETTE);
	}
		
	glColorTable(GL_RGB256, palette_address);
#endif
}

int tex_size(int size)
{
	switch (size)
	{
		case 8:
			size = TEXTURE_SIZE_8;
			break;
		case 16:
			size = TEXTURE_SIZE_16;
			break;
		case 32:
			size = TEXTURE_SIZE_32;
			break;
		case 64:
			size = TEXTURE_SIZE_64;
			break;
		case 128:
			size = TEXTURE_SIZE_128;
			break;
		case 256:
			size = TEXTURE_SIZE_256;
			break;
		case 512:
			size = TEXTURE_SIZE_512;
			break;
		case 1024:
			size = TEXTURE_SIZE_1024;
			break;
		default:
//			printf("invalid texture size, %d\n", size);
			size = -1;
			break;
	}
	return size;
}

int next_size_up(int size)
{
	if (size > 512)
		return 1024;
	if (size > 256)
		return 512;
	if (size > 128)
		return 256;
	if (size > 64)
		return 128;
	if (size > 32)
		return 64;
	if (size > 16)
		return 32;
	if (size > 8)
		return 16;
	if (size > 0)
		return 8;
	return 0;
}

unsigned int snapshot_activeTexture;
unsigned int *snapshot_nextBlock;
unsigned int snapshot_nextPBlock;
int snapshot_nameCount;
int snapshot_vram_used;

void ds_texture_snapshot(void)
{
#ifdef USE_3D
	extern unsigned int activeTexture;
	extern unsigned int *nextBlock;
	extern unsigned int nextPBlock;
	extern int nameCount;
	
	snapshot_activeTexture = activeTexture;
	snapshot_nextBlock = nextBlock;
	snapshot_nextPBlock = nextPBlock;
	snapshot_nameCount = nameCount;
	
	snapshot_vram_used = vram_used;
#endif
}

void ds_restore_snapshot(void)
{
#ifdef USE_3D
	extern unsigned int activeTexture;
	extern unsigned int *nextBlock;
	extern unsigned int nextPBlock;
	extern int nameCount;
	
	activeTexture = snapshot_activeTexture;
	nextBlock = snapshot_nextBlock;
	nextPBlock = snapshot_nextPBlock;
	nameCount = snapshot_nameCount;
	
	vram_used = snapshot_vram_used;
#endif
}

void ds_lcdswap(int sub_as_main)
{
	if (sub_as_main)
		lcdMainOnBottom();
	else
		lcdMainOnTop();
}

#define SSC_Disabled (0)
#define SSC_SDRAM (1)
#define SSC_CF (2)

#define SC_REG_UNLOCK	*(vu16*)(0x09FFFFFE)
static void SetSC_UNLOCK(int SSC)
{
  switch(SSC){
    case SSC_Disabled:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0001;
      SC_REG_UNLOCK = 0x0001;
      break;
    case SSC_SDRAM:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0005;
      SC_REG_UNLOCK = 0x0005;
      break;
    case SSC_CF:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0003;
      SC_REG_UNLOCK = 0x0003;
  }
}
#undef SC_REG_UNLOCK

void ds_exram_unlock(void)			//for ram access
{
#ifdef USE_EXTRA_RAM
	SetSC_UNLOCK(SSC_SDRAM);
#endif
}

void ds_exram_lock(void)			//for disk access
{
#ifdef USE_EXTRA_RAM
#ifdef ARM7_USES_EXRAM
	sysSetCartOwner(BUS_OWNER_ARM9);
#endif
	SetSC_UNLOCK(SSC_CF);
#endif
}

void ds_try_exram_unlock(void)
{
#ifdef USE_EXTRA_RAM
	ds_exram_unlock();
#ifdef ARM7_USES_EXRAM
	sysSetCartOwner(BUS_OWNER_ARM7);
#endif
#endif
}

void ds_try_exram_lock(void)
{
#ifdef USE_EXTRA_RAM
	ds_exram_lock();
#endif
}

void *ds_exram_base(void)
{
#ifdef USE_EXTRA_RAM
	return (void *)0x08000000;
#else
	return NULL;
#endif
}

unsigned int ds_exram_size(void)
{
#ifdef USE_EXTRA_RAM
	return 32 * 1024 * 1024;
#else
	return 0;
#endif
}

void ram_test(void)
{
#ifdef USE_EXTRA_RAM

	volatile unsigned short *ptr = (volatile unsigned short *)ds_exram_base();
	unsigned short special = 0x1234;
	unsigned short read_back = 0;
	
	int count = 0;
	while(1)
	{
		*ptr = special;
		read_back = *ptr;
		if (read_back != special)
			break;
		else
			count += 8;
		ptr += 8;
	}
	printf("%d words of goodness\n", count);
#endif
}

unsigned int ram_test_fast(void)
{
#ifdef USE_EXTRA_RAM

	volatile unsigned short *ptr = (volatile unsigned short *)ds_exram_base();
	unsigned short special = 0x1234;
	unsigned short read_back = 0;
	
	int count = 0;
	while(1)
	{
		*ptr = special;
		read_back = *ptr;
		if (read_back != special)
			break;
		else
			count += 512;
		ptr += 512;
	}
	return count * 2;
#endif
}

#define HINT_GET_READY 		1
#define HINT_RUNNING_OUT	2
#define HINT_FREE_TIME		3

void ds_hint_arm7(int hint)
{
	ipc_block_ready_9to7();
	
	switch(hint)
	{
		case HINT_GET_READY:
			quake_ipc_9to7->message_type = kGetReady;
			break;
		case HINT_RUNNING_OUT:
			quake_ipc_9to7->message_type = kRunningOut;
			break;
		case HINT_FREE_TIME:
			quake_ipc_9to7->message_type = kFreeTime;
			break;
		default:
			Sys_Error("invalid ARM7 hint, %d!\n", hint);
			break;
	}
	
	ipc_set_ready_9to7();
}

#define MALLOC_CHUNK_SIZE	256
#define MALLOC_CHUNK_SHIFT	8
#define NUM_MALLOC_CHUNKS	2048

#define VRAM_BASE 0x6800000

unsigned char chunks_in_use[NUM_MALLOC_CHUNKS];
unsigned char *chunk_base;

bool malloc_changed;

void vram_init(void)
{
	malloc_changed = false;
	memset(chunks_in_use, 0, sizeof(chunks_in_use));
	chunk_base = (unsigned char *)VRAM_BASE;
}

//FILE *vramfp = NULL;

void *vram_malloc(unsigned int size)
{
//	if (vramfp == NULL)
//	{
//		vramfp = fopen("vram.txt", "w");
//		if (!vramfp)
//			while(1);
//	}
	
	
	int slots_needed = size >> MALLOC_CHUNK_SHIFT;
	slots_needed += ((size & (MALLOC_CHUNK_SIZE - 1)) != 0);
	
//	printf("needs %d slots for %d bytes\n", slots_needed, size);
//	fprintf(vramfp, "needs %d slots for %d bytes\n", slots_needed, size);

	if ((size < 1) || (size > MALLOC_CHUNK_SIZE * NUM_MALLOC_CHUNKS))
	{
		printf("failed allocating %d bytes\n", size);
		while(1);
	}
	
	int found_slot = -1;
	
	for (int count = 0; count < NUM_MALLOC_CHUNKS; count++)
	{
		if (chunks_in_use[count] == 0)
		{
			int length = 0;
			for (int length_count = 0; length_count < NUM_MALLOC_CHUNKS - count; length_count++)
				if (chunks_in_use[count + length_count] == 0)
					length++;
				else
					break;
			
			if (length >= slots_needed)
			{
//				printf("found %d slots at %d\n", length, count);
				found_slot = count;
				break;
			}
		}
	}
	
	if (found_slot == -1)
	{
//		printf("didn\'t find a big enough slot\n");
//		fprintf(vramfp, "didn\'t find a big enough slot\n");
		return NULL;
	}
	else
	{
		for (int count = 1; count < slots_needed; count++)
			chunks_in_use[count + found_slot] = 255;
		
		if (slots_needed < 16)
			chunks_in_use[found_slot] = slots_needed & 0xf;
		else
		{
			chunks_in_use[found_slot] = (slots_needed & 0xf) | (1 << 7);
			chunks_in_use[found_slot + 1] = ((slots_needed & 0xf0) >> 4) | (1 << 7);
			chunks_in_use[found_slot + 2] = ((slots_needed & 0xf00) >> 8) | (1 << 7);
		}
		
//		printf("found slot at %08x\n", chunk_base + found_slot * MALLOC_CHUNK_SIZE);
//		fprintf(vramfp, "found slot at %08x\n", chunk_base + found_slot * MALLOC_CHUNK_SIZE);
		
		malloc_changed = true;
		return chunk_base + found_slot * MALLOC_CHUNK_SIZE;
	}
}

unsigned int vram_get_free_size(void)
{
	int block_count = 0;
	
	for (int count = 0; count < NUM_MALLOC_CHUNKS; count++)
		if (chunks_in_use[count] == 0)
			block_count++;
	
	return block_count * MALLOC_CHUNK_SIZE;
}

void vram_free(void *ptr)
{
	unsigned int int_ptr = (unsigned int)ptr;
//	printf("freeing %08x\n", (unsigned int)ptr);
//	while(1);
	
	if ((int_ptr < (unsigned int)chunk_base) || (int_ptr >= (unsigned int)chunk_base + MALLOC_CHUNK_SIZE * NUM_MALLOC_CHUNKS))
	{
		printf("trying to free an invalid address!\nout of range: %08x\n", int_ptr);
		while(1);
	}
	
	if ((int_ptr & (MALLOC_CHUNK_SIZE - 1)) != 0)
	{
		printf("trying to free an invalid address!\nnot on slot boundary: %08x\n", int_ptr);
		while(1);
	}
	
	int starting_slot = (int_ptr - (unsigned int)chunk_base) >> MALLOC_CHUNK_SHIFT;
	
	int slots_used = chunks_in_use[starting_slot];
	
	if (slots_used >> 7)
		slots_used = (slots_used & 0x7f)
					| ((chunks_in_use[starting_slot + 1] & 0x7f) << 4)
					| ((chunks_in_use[starting_slot + 2] & 0x7f) << 8);
	
//	printf("freeing %d, used %d slots, %d\n", starting_slot, slots_used, vram_get_free_size() >> MALLOC_CHUNK_SHIFT);
	
	if (slots_used == 255)
	{
		printf("trying to free data contents!\n");
		while(1);
	}
	
	if (slots_used == 0)
	{
		printf("double free to %08x\n", ptr);
		while(1);
	}
	
	chunks_in_use[starting_slot] = 0;
	
	for (int count = 1; count < slots_used; count++)
	{
//		if (chunks_in_use[count + starting_slot] != 255)
//		{
//			printf("allocating over used memory!\n");
//			while(1);
//		}
		chunks_in_use[count + starting_slot] = 0;
	}
	
	malloc_changed = true;
}

void copy_into_vram(unsigned int *dest, unsigned char *src, int size, int rows, int cols, int realX)
{
	if ((rows == 0) || (cols == 0))
		swiCopy((uint32*)src, dest , size / 4 | COPY_MODE_WORD);
	else
	{
		memset(dest, 0, size);
		
		realX = realX >> 2;
		
		for (int y = 0; y < rows; y++)
			memcpy(&dest[y * realX], &src[y * cols], cols);
	}
}

bool free_lru(int);

//from libnds, modified to do resizing on the fly
extern "C" uint32* getNextTextureSlot(int size);
extern "C" void glTexParameter(uint8 sizeX, uint8 sizeY, uint32* addr, int mode, uint32 param);
unsigned int glTexImage2DQuake(int target, int empty1, int type, int sizeX, int sizeY, int empty2, int param, uint8* texture,
	unsigned int rows, unsigned int cols, unsigned int realY, unsigned int realX) {
//---------------------------------------------------------------------------------
  uint32 size = 0;
  uint32* addr;
  uint32 vramTemp;

  size = 1 << (sizeX + sizeY + 6);
  

  switch (type) {
    case GL_RGB:
    case GL_RGBA:
      size = size << 1;
      break;
    case GL_RGB4:
      size = size >> 2;
      break;
    case GL_RGB16:
      size = size >> 1;
      break;

    default:
      break;
  }
  
//  addr = getNextTextureSlot(size);

	do
	{
//		printf("%d is free\n", vram_get_free_size());
		if ((rows == 0) && (cols == 0))
			addr = (unsigned int *)vram_malloc(size);
		else
			addr = (unsigned int *)vram_malloc(rows * realX);
		
		if (addr == NULL)
		{
			bool res = free_lru(super_framecount - 1);
			
			if (!res)
				return 0;
		}
	}
	while (addr == NULL);
	
	vram_used = NUM_MALLOC_CHUNKS * MALLOC_CHUNK_SIZE - vram_get_free_size();
	
//	if (texture == NULL);
//	printf("got %08x\n", addr);
  
  if(!addr)
  {
  	printf("failed to allocates texture memory\n");
  	return 0;
  }

  // unlock texture memory
  vramTemp = vramSetMainBanks(VRAM_A_LCD,VRAM_B_LCD,VRAM_C_LCD,VRAM_D_LCD);

  if (type == GL_RGB) {
    // We do GL_RGB as GL_RGBA, but we set each alpha bit to 1 during the copy
    u16 * src = (u16*)texture;
    u16 * dest = (u16*)addr;
    
    glTexParameter(sizeX, sizeY, addr, GL_RGBA, param);
    
    while (size--) {
      *dest++ = *src | (1 << 15);
      src++;
    }
  } else {
    // For everything else, we do a straight copy
    glTexParameter(sizeX, sizeY, addr, type, param);
    if (texture)
		copy_into_vram((unsigned int *)addr, (unsigned char *)texture, size, rows, cols, realX);
  }

  vramRestoreMainBanks(vramTemp);
  
//  printf("left %08x\n", addr);

  return (unsigned int)addr;
}

void ds_resize_in_place(int sizeX, int sizeY, unsigned char *image)
{
//	for (int count = 0; count < sizeX * sizeY; count += 2)
//	{
//		image[count >> 1] = image[count];
//	}
	
//	for (int y = 0; y < sizeY >> 1; y++)
//		memcpy(&image[y * (sizeX >> 1)], &image[(y * 2) * (sizeX >> 1)], sizeX >> 1);
//	memset(&image[(sizeX * sizeY) >> 1], 0, (sizeX * sizeY) >> 1);
	
	for (int y = 0; y < sizeY >> 1; y++)
		memcpy(&image[y * sizeX], &image[(y * 2) * sizeX], sizeX);
}

int fail_textures = 1;
int ds_teximage2d(int sizeX, int sizeY, unsigned char* texture, bool transparency, int transparent_colour)
{
//	if (fail_textures)
//		return -1;
#ifdef USE_3D
	int tex_xsize, tex_ysize;
	int resize_xsize, resize_ysize;
	int do_resize = 0;
	
//	for (int y = 0; y < sizeY >> 1; y++)
//		for (int x = 0; x < sizeX >> 1; x++)
//			texture[y * (sizeX >> 1) + x] = texture[y * 2 * sizeX + x * 2];
//			
//	sizeX = sizeX >> 1;
//	sizeY = sizeY >> 1;
//	
//	memset(texture + sizeX * sizeY, 0, sizeX * sizeY * 3);

//	for (int count = 0; count < (sizeX * sizeY) >> 1; count++)
//		texture[count] = texture[count * 2];
//			
//	sizeX = sizeX >> 1;
	
	if (transparency)
		for (int count = 0; count < sizeX * sizeY; count++)
				if (texture[count] == transparent_colour)
					texture[count] = 0;
	
	if ((sizeX == 0) || (sizeY == 0))
	{
		printf("zero size texture!\n");
		return -1;
	}
	if ((sizeX > 1024) || (sizeY > 1024))
	{
		printf("texture is too big! %dx%d\n", sizeX, sizeY);
		return -1;
	}
	
	tex_xsize = tex_size(sizeX);
	tex_ysize = tex_size(sizeY);
	resize_xsize = sizeX;
	resize_ysize = sizeY;
	
	if ((tex_xsize == -1) || (tex_ysize == -1))
	{
		//we're gonna have to enlarge it as it's not a normal size
		if (tex_xsize == -1)
			resize_xsize = next_size_up(sizeX);
		if (tex_ysize == -1)
			resize_ysize = next_size_up(sizeY);
//return -1;
//		if (resize_xsize * resize_ysize > 256 * 256)
//			return -1;
//		else
		{
//			int y;
//			memset((void *)resize, 0, resize_xsize * resize_ysize);
//			
//			for (y = 0; y < sizeY; y++)
//				memcpy(&resize[y * resize_xsize], &texture[y * sizeX], sizeX);
			
			do_resize = 1;
		}
	}
	
//	vram_used += resize_xsize * resize_ysize;
		
//#ifdef USE_DEBUGGER
//	char text_buf[100];
//	sprintf(text_buf, "%.2fk VRAM used (%.2fk), %.2fk total\n",
//		(float)(sizeX * sizeY) / 1024, (float)(resize_xsize * resize_ysize) / 1024, (float)vram_used / 1024);
////	DEBUG_MSG(text_buf);
//	printf(text_buf);
//#endif
	tex_xsize = tex_size(resize_xsize);
	tex_ysize = tex_size(resize_ysize);
	
	if (!do_resize)
		return glTexImage2DQuake(0, 0, GL_RGB256, tex_xsize, tex_ysize, 0, TEXGEN_TEXCOORD | GL_TEXTURE_WRAP_S | GL_TEXTURE_WRAP_T | (transparency ? GL_TEXTURE_COLOR0_TRANSPARENT : 0), texture, 0, 0, 0, 0) - 1;
//		return glTexImage2D(0, 0, GL_RGB256, tex_xsize, tex_ysize, 0, TEXGEN_TEXCOORD | GL_TEXTURE_WRAP_S | GL_TEXTURE_WRAP_T, texture) - 1;
	else
	{
//		printf("using resized version, %dx%d->%dx%d\n", sizeX, sizeY, resize_xsize, resize_ysize);
		return glTexImage2DQuake(0, 0, GL_RGB256, tex_xsize, tex_ysize, 0, TEXGEN_TEXCOORD | GL_TEXTURE_WRAP_S | GL_TEXTURE_WRAP_T | (transparency ? GL_TEXTURE_COLOR0_TRANSPARENT : 0), texture, sizeY, sizeX, resize_ysize, resize_xsize) - 1;
	}
#endif
	return 0;
}

#define MAX_MANAGED_TEXTURES 400

#define TEXTURE_VALID			1
#define TEXTURE_BSP				2
#define TEXTURE_IS_LOADED		4
#define TEXTURE_NEEDS_LOADING	8
#define TEXTURE_NEVER_UNLOAD	16
#define TEXTURE_IS_TRANSPARENT	32

#define TEXTURE_VALID_MASK			0xfffe
#define TEXTURE_BSP_MASK			0xfffd
#define TEXTURE_IS_LOADED_MASK		0xfffb
#define TEXTURE_NEEDS_LOADING_MASK	0xfff7
#define TEXTURE_NEVER_UNLOAD_MASK	0xffef

struct managed_texture
{
	unsigned short handle;
#ifdef TEXTURES_HAVE_NAMES
	char name[15];
#endif
//	bool valid;
//	bool bsp;
//	
//	bool is_loaded;
//	bool needs_loading;
//	bool never_unload;
	unsigned short mask_data;
	
	unsigned int last_used;
	void *address;
	
//	int loaded_rows;
	
	unsigned short fp;
	unsigned int seek_pos;
	
	unsigned short sizeX, sizeY;
	unsigned short half_size;

//	unsigned short transparency;
	unsigned short transparent_colour;
};
#ifdef TEXTURE_CACHE_IN_VRAM
managed_texture *texture_cache = NULL;
#else
managed_texture *texture_cache[MAX_MANAGED_TEXTURES];
#endif

int num_managed_textures = 0;

void init_textures(void)
{
#ifdef TEXTURE_CACHE_IN_VRAM
	vramSetBankI(VRAM_I_LCD);
	texture_cache = (managed_texture *)0x68A0000;
#endif
	num_managed_textures = 0;
//	memset(texture_cache, 0, sizeof(managed_texture) * MAX_MANAGED_TEXTURES);

	if (sizeof(managed_texture) * MAX_MANAGED_TEXTURES > 16384)
		Sys_Error("texture cache will not fit in memory! size %d\n", sizeof(managed_texture) * MAX_MANAGED_TEXTURES);
	
	unsigned short *ptr = (unsigned short *)texture_cache;
	for (int count = 0; count < (sizeof(managed_texture) * MAX_MANAGED_TEXTURES) >> 1; count++)
		ptr[count] = 0;
}

int register_texture(bool bsp, char *name, int fp, int seek, int sizeX, int sizeY, bool transparency, int transparent_colour, int half_width, int half_height)
{
	//test to see if it's already been loaded once
#ifdef TEXTURES_HAVE_NAMES
	if (!bsp)
		for (int count = 0; count < num_managed_textures; count++)
			if (strncmp(name, texture_cache[count].name, 15) == 0)
				return count;
#else
	if (!bsp)
		for (int count = 0; count < num_managed_textures; count++)
			if ((texture_cache[count].fp == fp) && (texture_cache[count].seek_pos == seek))
				return count;
#endif	
//	printf("28 is %s\n", texture_cache[28].name);
//	printf("29 is %s\n", texture_cache[29].name);

//	if (fail_textures)
//		return -1;

	if (fp < 0)
		return -1;
	
	int id = num_managed_textures;
	
	for (int count = 0; count < num_managed_textures; count++)
		if (!(texture_cache[count].mask_data & TEXTURE_VALID))
		{
//			printf("replacing %d\n", count);
			id = count;
			break;
		}
	
	if (id == num_managed_textures)
		num_managed_textures++;
	
	if (id >= MAX_MANAGED_TEXTURES)
	{
		printf("ran out of texture handles!\n");
		printf("trying to make a texture for %s\n", name);
		while(1);
	}
	
	struct managed_texture *t = &texture_cache[id];
	
	t->handle = ds_gentexture();
#ifdef TEXTURES_HAVE_NAMES
	memset(t->name, 0, 15);
	
	for (int count = 0; count < 14; count++)
		t->name[count] = name[count];
	
	t->name[14] = 0;
#endif

	t->mask_data = 0;
	
	if (bsp)
		t->mask_data |= TEXTURE_BSP;
	t->mask_data |= TEXTURE_VALID;
	
//	t->bsp = bsp;
//	t->valid = true;
//	t->is_loaded = false;
//	t->needs_loading = false;
//	t->never_unload = false;
	t->last_used = 0;
	t->fp = fp;
	t->seek_pos = seek;
	t->sizeX = sizeX;
	t->sizeY = sizeY;
//	t->transparency = transparency;

	if (transparency)
		t->mask_data |= TEXTURE_IS_TRANSPARENT;

	t->transparent_colour = transparent_colour;
	t->half_size = (half_width << 1) | half_height;
	t->address = NULL;
	
//	printf("bound %d: %s, size %d\n", num_managed_textures - 1, name, sizeX * sizeY);
	
	return id;
}

//int fail_after = 0;

int register_texture_deferred(char *name, unsigned char *data, int sizeX, int sizeY, int transparency, int transparent_colour)
{
//	//test to see if it's already been loaded once
//	for (int count = 0; count < num_managed_textures; count++)
//		if (strcmp(name, texture_cache[count].name) == 0)
//			return count;

//	if (fail_textures)
//		return -1;
	
	int id = num_managed_textures;
	num_managed_textures++;
	
	for (int count = 0; count < num_managed_textures; count++)
		if (!(texture_cache[count].mask_data & TEXTURE_VALID))
		{
			id = count;
			num_managed_textures--;
			break;
		}
	
	if (id >= MAX_MANAGED_TEXTURES)
	{
		printf("ran out of texture handles!\n");
		printf("trying to make a texture for %s\n", name);
		while(1);
	}
	
//	printf("loading %s as %d DEFERRED\n", name, id);
	
	num_managed_textures++;
	
	struct managed_texture *t = &texture_cache[id];
	
	t->handle = ds_gentexture();
	
#ifdef TEXTURES_HAVE_NAMES
	strncpy(t->name, name, 15);
	t->name[14] = NULL;
#endif

	t->mask_data = 0;
	t->mask_data |= TEXTURE_VALID;
	
//	t->valid = true;
//	t->bsp = false;
//	t->is_loaded = false;
//	t->needs_loading = false;
//	t->never_unload = false;
	t->last_used = 0;
	t->fp = NULL;
	t->seek_pos = (unsigned int)data;
	t->sizeX = sizeX;
	t->sizeY = sizeY;
//	t->transparency = transparency;

	if (transparency)
		t->mask_data |= TEXTURE_IS_TRANSPARENT;

	t->transparent_colour = transparent_colour;
	t->half_size = 0;
	
	t->address = NULL;
	
//	printf("bound %d: %s, size %d\n", num_managed_textures - 1, name, sizeX * sizeY);
	
	return id;
}

int register_texture_immediate(char *name, unsigned char *data, int sizeX, int sizeY, int transparency, int transparent_colour)
{
	Sys_Error("registering immediate texture\n");
////	//test to see if it's already been loaded once
////	for (int count = 0; count < num_managed_textures; count++)
////		if (strcmp(name, texture_cache[count].name) == 0)
////			return count;
//
////	if (fail_textures)
////		return -1;
//	
//	int id = num_managed_textures;
//	num_managed_textures++;
//	
//	for (int count = 0; count < num_managed_textures; count++)
//		if (!texture_cache[count].valid)
//		{
//			id = count;
//			num_managed_textures--;
//			break;
//		}
//	
//	if (id >= MAX_MANAGED_TEXTURES)
//	{
//		printf("ran out of texture handles!\n");
//		printf("trying to make a texture for %s\n", name);
//		while(1);
//	}
//	
////	printf("loading %s as %d\n", name, id);
//	
//	num_managed_textures++;
//	
//	struct managed_texture *t = &texture_cache[id];
//	
//	t->handle = ds_gentexture();
//	
//	strncpy(t->name, name, 15);
//	t->name[14] = NULL;
//	
//	t->valid = true;
//	t->bsp = false;
//	t->is_loaded = true;
//	t->needs_loading = false;
//	t->never_unload = true;
//	t->last_used = 0;
//	t->fp = NULL;
//	t->seek_pos = 0;
//	t->sizeX = sizeX;
//	t->sizeY = sizeY;
//	t->transparency = transparency;
//	t->transparent_colour = transparent_colour;
//	t->half_size = 0;
//	
////	printf("trans %d, %d\n", transparency, transparent_colour);
////	printf("size %dx%d\n", sizeX, sizeY);
//
////	if (fail_after > 173)
////	{
////		fclose(vramfp);
////		while(1);
////	}
////		return -1;
//	
//	ds_bindtexture(t->handle);
//	unsigned int addr = ds_teximage2d(sizeX, sizeY, data, transparency, transparent_colour);
//	if (addr == -1)
//		return -1;
//		
////	printf("addr is %08x, to %08x\n", addr, addr + sizeX * sizeY);
//	
////	if (strcmp(name, "lower sky") == 0)
////		while(1);
//	
////	fail_after++;
//		
//	t->address = (void *)addr;
//	
//	return id;
}

void *get_texture_address(int id)
{
	return texture_cache[id].address;
}

unsigned short dummy[512];
int into_cache = -1;
bool load_texture(int id)
{
	unsigned short *vram_e_base = (unsigned short *)0x6880000;
	struct managed_texture *t = &texture_cache[id];
	
	if (into_cache == -1)
	{
//#ifdef TEXTURES_HAVE_NAMES
//		printf("loading %s into cache\n", t->name);
//#else
//		printf("loading into cache\n");
//#endif
//		printf("size %dx%d\n", t->sizeX, t->sizeY);

		int rows = t->sizeY;
		int cols = t->sizeX;
		int big_cols = next_size_up(cols);
		
		if (t->fp)
			Sys_FileSeek(t->fp, t->seek_pos);
		
		if (t->half_size & 0x1)
		{
			if (t->mask_data & TEXTURE_IS_TRANSPARENT)
			{
				for (int y = 0; y < rows; y++)
				{
					Sys_FileRead(t->fp, dummy, cols);
					
					unsigned char *src = (unsigned char *)dummy;
					unsigned char *dest = (unsigned char *)dummy;
					
					for (int count = 0; count < cols; count++)
					{
						unsigned char b1 = src[count];
						
						if (b1 == t->transparent_colour)
							b1 = 0;
						
						dest[count] = b1;
					}
		
					unsigned short *ptr = (unsigned short *)&((unsigned char *)vram_e_base)[y * big_cols];
					for (int count = 0; count < (cols >> 1); count++)
						ptr[count] = dummy[count];
					
					Sys_FileRead(t->fp, dummy, cols);
				}
			}
			else
				for (int y = 0; y < rows; y++)
				{
					Sys_FileRead(t->fp, dummy, cols);
		
					unsigned short *ptr = (unsigned short *)&((unsigned char *)vram_e_base)[y * big_cols];
					for (int count = 0; count < (cols >> 1); count++)
						ptr[count] = dummy[count];
					
					Sys_FileRead(t->fp, dummy, cols);
				}
		}
		else
			for (int y = 0; y < rows; y++)
			{
				if (t->fp)
				{
					Sys_FileRead(t->fp, dummy, cols);
					
//					if (t->transparency)
					if (t->mask_data & TEXTURE_IS_TRANSPARENT)
					{
						unsigned char *src = (unsigned char *)dummy;
						unsigned char *dest = (unsigned char *)dummy;
						
						for (int count = 0; count < cols; count++)
						{
							unsigned char b1 = src[count];
							
							if (b1 == t->transparent_colour)
								b1 = 0;
							
							dest[count] = b1;
						}
					}
				}
				else
				{
					unsigned char *src = (unsigned char *)(t->seek_pos + y * cols);
					
//					if (t->transparency)
					if (t->mask_data & TEXTURE_IS_TRANSPARENT)
					{
						for (int count = 0; count < cols; count += 2)
						{
							unsigned char b1 = src[count];
							unsigned char b2 = src[count + 1];
							
							if (b1 == t->transparent_colour)
								b1 = 0;
							if (b2 == t->transparent_colour)
								b2 = 0;
							
							dummy[count >> 1] = (b2 << 8) | b1;
						}
					}
					else
						memcpy(dummy, src, cols); 
				}

				unsigned short *ptr = (unsigned short *)&((unsigned char *)vram_e_base)[y * big_cols];
				for (int count = 0; count < (cols >> 1); count++)
					ptr[count] = dummy[count];
			}
		
		into_cache = id;
	}
	
	if (into_cache == id)
	{
		unsigned short vcount = *(volatile unsigned short *)0x4000006;
		if ((vcount < HBLANK_START_LOAD) || (vcount >= HBLANK_END_LOAD))
		{
//			printf("falling out, vcount %d\n", vcount);
			return true;
		}
		
		ds_bindtexture(t->handle);
		unsigned int addr = ds_teximage2d(t->sizeX, t->sizeY, NULL,
			(t->mask_data & TEXTURE_IS_TRANSPARENT) == TEXTURE_IS_TRANSPARENT,
			t->transparent_colour);
		t->address = (void *)(addr + 1);	//yeah
		
//#ifdef TEXTURES_HAVE_NAMES
//		printf("loading %s from cache, %d\n", t->name, *(volatile unsigned short *)0x4000006);
//#else
//		printf("loading from cache\n");
//#endif
		
//		printf("texture address is %08x\n", t->address);

		if (t->address)
		{
			int rows = t->sizeY;
			int cols = t->sizeX;
			int big_cols = next_size_up(cols);
			
//			unsigned short marker_1 = *(volatile unsigned short *)0x4000006;
			
			ds_unlock_vram();
			
			unsigned short *dest = (unsigned short *)t->address;
			unsigned short *src = vram_e_base;
			
//			for (int count = 0; count < big_cols * rows; count += 2)
//			{
//				*dest = *src;
//				dest++;
//				src++;
//			}
			dmaCopyHalfWords(0, src, dest, big_cols * rows);
				
			ds_lock_vram();
			
//			unsigned short marker_2 = *(volatile unsigned short *)0x4000006;
			
//			printf("loaded %d b, %d\n", big_cols * rows, marker_2 - marker_1);
			
	//		t->is_loaded = true;
	//		t->needs_loading = false;
			t->mask_data |= TEXTURE_IS_LOADED;
		}
		t->mask_data = t->mask_data & TEXTURE_NEEDS_LOADING_MASK;
		
		into_cache = -1;
	}
	else
	{
//		printf("%d is in cache\n", into_cache);
		t->mask_data = t->mask_data & TEXTURE_NEEDS_LOADING_MASK;
	}
	return false;
}

//bool load_texture(int id)
//{
//	struct managed_texture *t = &texture_cache[id];
//	
//	printf("loading %s\n", t->name);
//	
//	ds_bindtexture(t->handle);
//	unsigned int addr = ds_teximage2d(t->sizeX, t->sizeY, NULL, t->transparency, t->transparent_colour);
//	t->address = (void *)(addr + 1);	//yeah
//	
//	printf("texture address is %08x\n", t->address);
//	
//	bool outa_time = false;
//	
//	if (t->address)
//	{	
//		ds_unlock_vram();
//		
//		int rows = t->sizeY;
//		int cols = t->sizeX;
//		int big_cols = next_size_up(cols);
//		
////		if (*(volatile unsigned short *)0x4000006 > 212) { return true; }
//		
//		Sys_FileSeek(t->fp, t->seek_pos + cols * t->loaded_rows);
//		
////		if (*(volatile unsigned short *)0x4000006 > 212) { return true; }
//		
//		t->loaded_rows = 0;
//		
//		int read = 0;
//
//		if (t->half_size & 0x1)
//			for (int y = 0; y < rows; y++)
//			{
////				if (*(volatile unsigned short *)0x4000006 > 212) { outa_time = true; t->loaded_rows = y; break; }
//				
//				Sys_FileRead(t->fp, dummy, cols);
//	
//				unsigned short *ptr = (unsigned short *)&((unsigned char *)t->address)[y * big_cols];
//				
////				if (*(volatile unsigned short *)0x4000006 > 212) { outa_time = true; t->loaded_rows = y; break; }
//				
//				for (int count = 0; count < (cols >> 1); count++)
//					ptr[count] = dummy[count];
//				
////				if (*(volatile unsigned short *)0x4000006 > 212) { outa_time = true; t->loaded_rows = y + 1; break; }
//				
//				Sys_FileRead(t->fp, dummy, cols);
//			}
//		else
//			for (int y = 0; y < rows; y++)
//			{
////				if (*(volatile unsigned short *)0x4000006 > 212) { outa_time = true; t->loaded_rows = y; break; }
//				
//				Sys_FileRead(t->fp, dummy, cols);
//				
////				if (*(volatile unsigned short *)0x4000006 > 212) { outa_time = true; t->loaded_rows = y; break; }
//	
//				unsigned short *ptr = (unsigned short *)&((unsigned char *)t->address)[y * big_cols];
//				for (int count = 0; count < (cols >> 1); count++)
//					ptr[count] = dummy[count];
//				
////				if (*(volatile unsigned short *)0x4000006 > 212) { outa_time = true; t->loaded_rows = y + 1; break; }
//			}
//		
////		unsigned short *ptrs = (unsigned short *)&((unsigned char *)t->address)[0];
////		ptrs[0] = 0;
////		ptrs[1] = 0;
//		
//		ds_lock_vram();
//		
//		t->is_loaded = true;
//		
////		printf("loading %s\n", t->name);
//	}
//	
//	t->needs_loading = false;
//	
//	return outa_time;
//}

volatile int texture_critical_section = 0;
int textures_needing_load = 0;

void load_textures(void)
{
	if (!texture_critical_section)
	{
//		printf("%d textures need load\n", textures_needing_load);
		if (textures_needing_load)
		{
			while (*((volatile unsigned short *)0x4000006) > HBLANK_START_LATE_LOAD);
			while (*((volatile unsigned short *)0x4000006) < HBLANK_START_LOAD);
			
//			printf("%d vram free before loading\n", vram_get_free_size());
			texture_critical_section = 1;
				
			for (int count = 0; count < num_managed_textures; count++)
				if (texture_cache[count].mask_data & TEXTURE_NEEDS_LOADING)
					if (load_texture(count))	//outatime
						break;
			
			texture_critical_section = 0;
//			printf("%d vram free after loading\n", vram_get_free_size());
		}
		textures_needing_load = 0;
	}
}

//void unload_textures_2(void)
//{
//	printf("deleting all textures\n");
//	if (!texture_critical_section)
//	{
//		texture_critical_section = 1;
//		for (int count = 0; count < num_managed_textures; count++)
//		{
//			if (texture_cache[count].mask_data & TEXTURE_VALID)
//			{
//				if (texture_cache[count].mask_data& TEXTURE_IS_LOADED)
//				{
////					printf("freeing %s, %dx%d, %08x\n",
////						texture_cache[count].name,
////						texture_cache[count].sizeX,
////						texture_cache[count].sizeY,
////						texture_cache[count].address);
//					vram_free(texture_cache[count].address);
//					
//					texture_cache[count].address = NULL;
//					texture_cache[count].mask_data = texture_cache[count].mask_data & TEXTURE_IS_LOADED_MASK;
//					texture_cache[count].last_used = 0;
//				}
//			}
//		}
//
//		texture_critical_section = 0;
//		
//		printf("%d bytes vram free\n", vram_get_free_size());
//		
//		if (vram_get_free_size() < 524288)
//			while(1);
//	}
//	else
//	{
//		Sys_Error("trying to free textures during critical section!\n");
//	}
//}

void unload_textures(void)
{
	if (!texture_critical_section)
	{
		texture_critical_section = 1;
		into_cache = -1;
		
		for (int count = 0; count < num_managed_textures; count++)
		{
			if (/*(texture_cache[count].mask_data & TEXTURE_BSP) && */(texture_cache[count].mask_data & TEXTURE_VALID))
			{
				if (texture_cache[count].mask_data & TEXTURE_IS_LOADED)
				{
//					printf("freeing %s, %dx%d, %08x\n",
//						texture_cache[count].name,
//						texture_cache[count].sizeX,
//						texture_cache[count].sizeY,
//						texture_cache[count].address);
					vram_free(texture_cache[count].address);
					texture_cache[count].address = NULL;
				}
//				texture_cache[count].valid = false;
				if (texture_cache[count].mask_data & TEXTURE_BSP)
				{
					texture_cache[count].mask_data = texture_cache[count].mask_data & TEXTURE_VALID_MASK;
					texture_cache[count].mask_data = texture_cache[count].mask_data & TEXTURE_BSP_MASK;
				}
					
				texture_cache[count].mask_data = texture_cache[count].mask_data & TEXTURE_IS_LOADED_MASK;
				texture_cache[count].last_used = 0;
			}
		}
		
		printf("leaked %d bytes of VRAM\n", 512 * 1024 - vram_get_free_size());

		texture_critical_section = 0;
		
//		unload_textures_2();
	}
	else
	{
		Sys_Error("trying to free textures during critical section!\n");
	}
}

int find_texture(char *name)
{
#ifdef TEXTURES_HAVE_NAMES
	for (int count = 0; count < num_managed_textures; count++)
		if (strcmp(texture_cache[count].name, name) == 0)
			return count;
#endif
	
	return -1;
}

int find_texture(void *addr)
{
	for (int count = 0; count < num_managed_textures; count++)
		if ((unsigned int)texture_cache[count].seek_pos == (unsigned int)addr) 
			return count;
	
	return -1;
}

bool free_lru(int min_time)
{
	int lru = -1;
	unsigned int lru_time = -1;
	
//	printf("free lru\n");
	
	for (int count = 0; count < num_managed_textures; count++)
//		if ((texture_cache[count].last_used < lru_time)
//			&& (texture_cache[count].last_used < min_time)
//			&& texture_cache[count].is_loaded
//			&& !texture_cache[count].never_unload
//			&& texture_cache[count].valid)
		if ((texture_cache[count].last_used < lru_time)
//			&& (texture_cache[count].last_used < min_time)
			&& (texture_cache[count].mask_data & TEXTURE_IS_LOADED)
			&& !(texture_cache[count].mask_data & TEXTURE_NEVER_UNLOAD)
			&& (texture_cache[count].mask_data & TEXTURE_VALID))
		{
			lru_time = texture_cache[count].last_used;
			lru = count;
		}
	
//	printf("lru is %d\n", lru);
//	printf("freeing %s\n", texture_cache[lru].name);
	
	if (lru == -1)
	{
		printf("no more textures to free!\n");
		return false;
	}
	
	texture_cache[lru].mask_data = texture_cache[lru].mask_data & TEXTURE_IS_LOADED_MASK;
//	texture_cache[lru].is_loaded = false;
	texture_cache[lru].last_used = 0;
	
	vram_free(texture_cache[lru].address);
	texture_cache[lru].address = NULL;
	
	return true;
}

void dump_texture_state(void)
{
		FILE *fp;
		char filename[50];
		sprintf(filename, "/textures_%d.txt", super_framecount);
		fp = fopen(filename, "w");
		
		fprintf(fp, "hblanks is %d, blanks is %d, sfc is %d\n", hblanks, blanks, super_framecount);
		fprintf(fp, "%d managed textures\n", num_managed_textures);
		fprintf(fp, "%d bytes free\n", vram_get_free_size());
		
		for (int count = 0; count < num_managed_textures; count++)
			fprintf(fp, "%d: name %s, bsp %d, valid %d, loaded %d, needs loading %d, never unload %d, address %08x, lru %d, size %dx%d\n",
				count,
#ifdef TEXTURES_HAVE_NAMES
				texture_cache[count].name,
#else
				"no name",
#endif
				(texture_cache[count].mask_data & TEXTURE_BSP) > 0,
				(texture_cache[count].mask_data & TEXTURE_VALID) > 0,
				(texture_cache[count].mask_data & TEXTURE_IS_LOADED) > 0,
				(texture_cache[count].mask_data & TEXTURE_NEEDS_LOADING) > 0,
				(texture_cache[count].mask_data & TEXTURE_NEVER_UNLOAD) > 0,
				texture_cache[count].address,
				texture_cache[count].last_used,
				texture_cache[count].sizeX, texture_cache[count].sizeY);
		
		fclose(fp);
		printf("dumped texture state\n");
		
//		unload_textures_2();

//		while(1);	
}

bool bind_texture(int id)
{
	struct managed_texture *t = &texture_cache[id];
	bool bound = false;
	
	if (id >= num_managed_textures)
	{
		printf("trying to bind unknown texture! (%d)\n", id);
		register unsigned int lr asm("lr");
		printf("called from %08x\n", lr);
		while(1);
	}
	
//	if ((t->fp == NULL) && (t->never_unload == false))
//		printf("binding deferred %d\n", id);
	
	if (id >= 0)
	{
//		printf("trying to bind %d (%d), %d bytes free\n", id, t->handle, vram_get_free_size());
		
//		if (t->is_loaded)
		if (t->mask_data & TEXTURE_IS_LOADED)
		{
			ds_bindtexture(t->handle);
			t->last_used = super_framecount;
			bound = true;
		}
		else
		{
			ds_bindtexture(0);
//			t->needs_loading = true;
			t->mask_data |= TEXTURE_NEEDS_LOADING;
			
			textures_needing_load++;
		}
	}
	else
		ds_bindtexture(0);
	
	return bound;
}

extern "C" void ds_soundsetup(unsigned char vol, unsigned char pan)
{
	setGenericSound(11025, vol, pan, 1);
}

extern "C" void ds_playsound(void *data, int length)
{
	playGenericSound(data, length);
}

void ds_stopallsounds(void)
{
	printf("ARM9: Stopping sounds\n");
	ipc_block_ready_9to7();
	
	quake_ipc_9to7->message_type = kStopAllSounds;
	ipc_set_ready_9to7();
	
	ipc_block_ready_9to7();
}

void ds_reinit_console(unsigned char *font)
{
#ifndef USE_DEBUGGER
	printf("Changing console font\n");
	unsigned short *ptr = (u16*)CHAR_BASE_BLOCK_SUB(0);
	extern unsigned short *d_8to16table;
 

	for (int new_char = 0; new_char < 255; new_char++)
	{
		int row = new_char >> 4;
		int col = new_char & 15;
		unsigned char *source = font + (row << 10) + (col << 3);
		
		for (int count = 16 * new_char; count < 17 * new_char; count += 2)
		{
			unsigned short new_line = 0;

//			if (source[0])
//				new_line = 0xf;
//			if (source[1])
//				new_line |= 0xf0;
//			if (source[2])
//				new_line |= 0xf00;
//			if (source[3])
//				new_line |= 0xf000;
				
			ptr[count * 2] = source[0] | (source[1] << 8);
			ptr[count * 2 + 1] = source[2] | (source[3] << 8);
			ptr[count * 2 + 2] = source[4] | (source[5] << 8);
			ptr[count * 2 + 3] = source[6] | (source[7] << 8);
				
//			ptr[count + 1] = new_line;
			
//			new_line = 0;
				
//			if (source[4])
//				new_line = 0xf;
//			if (source[5])
//				new_line |= 0xf0;
//			if (source[6])
//				new_line |= 0xf00;
//			if (source[7])
//				new_line |= 0xf000;
//			new_line = 0;
//			ptr[count+1] = new_line;
			
			source += 128;
		}
	}
#endif
}

#ifdef USE_WIFI
#define WIFI_REFRESH_TIMER 50
void WifiTimer(void)
{
   Wifi_Timer(WIFI_REFRESH_TIMER);
}

void arm9_synctoarm7()
{ // send fifo message
   REG_IPC_FIFO_TX=0x87654321;
}

// interrupt handler to receive fifo messages from arm7
void arm9_fifo()
{ // check incoming fifo messages
   u32 value = REG_IPC_FIFO_RX;
   if(value == 0x87654321) Wifi_Sync();
}

void wifi_go(void)
{
	printf("wireless init...........");
	
	REG_IPC_FIFO_CR = IPC_FIFO_ENABLE | IPC_FIFO_SEND_CLEAR; // enable & clear FIFO
	
	u32 Wifi_pass= Wifi_Init(WIFIINIT_OPTION_USELED);
	REG_IPC_FIFO_TX=0x12345678;
	REG_IPC_FIFO_TX=Wifi_pass;
	
	*((volatile u16 *)0x0400010E) = 0; // disable timer3
	
	irqInit();
	
	irqSet(IRQ_TIMER3, WifiTimer); // setup timer IRQ
	irqEnable(IRQ_TIMER3);
	irqSet(IRQ_FIFO_NOT_EMPTY, arm9_fifo); // setup fifo IRQ
	irqEnable(IRQ_FIFO_NOT_EMPTY);
	
	REG_IPC_FIFO_CR = IPC_FIFO_ENABLE | IPC_FIFO_RECV_IRQ; // enable FIFO IRQ

	Wifi_SetSyncHandler(arm9_synctoarm7); // tell wifi lib to use our handler to notify arm7

	TIMER_DATA(3) = TIMER_FREQ_256(WIFI_REFRESH_TIMER);
 	TIMER_CR(3) = TIMER_ENABLE | TIMER_IRQ_REQ | TIMER_DIV_256;

	while(Wifi_CheckInit()==0) { // wait for arm7 to be initted successfully
		while((*((u16 volatile *) 0x04000006)) > 192); // wait for vblank
		while((*((u16 volatile *) 0x04000006)) < 192);
	}
	
	printf("success!");
	
	printf("Connecting to AP........");
	
	int i;
	Wifi_AutoConnect(); // request connect
	while(1) {
		i=Wifi_AssocStatus(); // check status
		if(i==ASSOCSTATUS_ASSOCIATED) {
			printf("success!");
			break;
		}
		if(i==ASSOCSTATUS_CANNOTCONNECT) {
			Sys_Error("failure!");
			break;
		}
	}
}
#endif

void ds_handle_sleep(void)
{
	printf("ARM9: going to sleep\n");
	
	irqDisable((IRQ_MASK)(IRQ_VBLANK | IRQ_HBLANK));
#ifdef IPC_IN_TIMER
	irqDisable(IRQ_TIMER0);
#endif
	
	REG_IPC_FIFO_TX = 0xabadbabe;
	for (volatile int wait_count = 0; wait_count < 200000; wait_count++);
	REG_IPC_FIFO_CR = REG_IPC_FIFO_CR | (1 << 3);
	
	powerOFF(POWER_ALL);
	
	while (REG_IPC_FIFO_CR&IPC_FIFO_RECV_EMPTY);
		
	unsigned int data = REG_IPC_FIFO_RX;
	
	powerON(POWER_ALL);
	irqEnable((IRQ_MASK)(IRQ_VBLANK | IRQ_HBLANK));
#ifdef IPC_IN_TIMER
	irqEnable(IRQ_TIMER0);
#endif
	
	if (data != 0xabadbabe)
	{
		printf("woke with the wrong message, %08x\n", data);
		while(1);
	}
	
//	REG_IPC_FIFO_TX = 0xabadbabe;
//	for (volatile int wait_count = 0; wait_count < 200000; wait_count++);
//	REG_IPC_FIFO_CR = REG_IPC_FIFO_CR | (1 << 3);
	
	printf("ARM9: awake\n");
}

int real_main(int argc, unsigned char **argv)
{
	powerON(POWER_ALL);
	
	REG_IPC_FIFO_CR = IPC_FIFO_ENABLE | IPC_FIFO_SEND_CLEAR;
  
#ifndef USE_3D
	// Use the main screen for output
 	videoSetMode(MODE_FB0 | DISPLAY_BG0_ACTIVE);
 	vramSetBankA(VRAM_A_LCD);
#else
    
//    videoSetMode(MODE_0_2D | DISPLAY_BG0_ACTIVE);
// 	vramSetBankA(VRAM_A_MAIN_BG);
 	
 	BG0_CR = BG_MAP_BASE(31);
 	videoSetMode(MODE_0_3D);
 	
 	vramSetBankA(VRAM_A_TEXTURE);
 	vramSetBankB(VRAM_B_TEXTURE);
 	vramSetBankC(VRAM_C_TEXTURE);
 	vramSetBankD(VRAM_D_TEXTURE);
//	vramSetBankC(VRAM_C_ARM7);
//	vramSetBankD(VRAM_D_ARM7);
#endif
	
#ifdef USE_QUAKE_CONSOLE
	videoSetModeSub(MODE_3_2D | DISPLAY_BG3_ACTIVE);
	vramSetBankH(VRAM_H_SUB_BG);
	SUB_BG3_CR = BG_BMP16_128x128 | BG_BMP_BASE(0);
	
	SUB_BG3_XDX = 1<<7;
 	SUB_BG3_YDY = 1<<7;
 	SUB_BG3_XDY = 0;
 	SUB_BG3_YDX = 0;
 	SUB_BG3_CX = 0;
 	SUB_BG3_CY = 0;
 #else
	videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
	vramSetBankH(VRAM_H_SUB_BG);
	SUB_BG0_CR = BG_MAP_BASE(15) | BG_256_COLOR;
		
	BG_PALETTE_SUB[255] = RGB15(31,31,31);
	
	consoleInitDefault((u16*)SCREEN_BASE_BLOCK_SUB(15), (u16*)CHAR_BASE_BLOCK_SUB(0), 8);
 #endif
 
 	vramSetBankE(VRAM_E_LCD);
 	
 	vram_init();
	init_textures();
 
	glViewPort(0,0,255,191);
	glClearColor(0,0,0);

	glClearDepth(0x7FFF);
	
//	lcdSwap();

	printf("Starting up\n");
	
#if !(defined USE_DEBUGGER) && !(defined USE_WIFI)
	irqInit();
#endif
	printf("fat init...");
	if (fatInitDefault())
		printf("ok\n");
	else
	{
		printf("failed!\n");
		while(1);
	}

#ifdef USE_DEBUGGER
	printf("starting debugger\n");
	//start up the wireless, connect to the AP
	set_verbosity(VERBOSE_ERROR | VERBOSE_INFO/* | VERBOSE_TRACE | VERBOSE_MMAP*/);
	wireless_init(1);
	wireless_connect();

	printf("connecting...");
	//connect to the PC stub, initialise the debugger
	debugger_connect_tcp(192, 168, 1, 10);
	debugger_init();					//from here on in you can get the debugger's attention
	printf("ok\n");
	
	//this will check for new commands from the debugger - call this repeatedly
	//the first time it is call it will break until you decide to resume
	//you can call this more than once in your program
	user_debugger_update();			//the earlier you call it, the earlier you can pause your program
#endif

#ifdef USE_WIFI
	printf("starting wireless\n");
	wifi_go();
#endif
	
	irqSet(IRQ_VBLANK, vblank_handler);
	irqEnable(IRQ_VBLANK);
	irqSet(IRQ_HBLANK, hblank_handler);
	irqEnable(IRQ_HBLANK);
	
#ifdef IPC_IN_TIMER
	irqSet(IRQ_TIMER0, handle_ipc);
	TIMER_DATA(0) = TIMER_FREQ_256(10 * 60);
 	TIMER_CR(0) = TIMER_ENABLE | TIMER_IRQ_REQ | TIMER_DIV_256;
#endif

#ifdef DO_FP_TESTS
	do_fp_tests();
#endif

	FILE *fp = fopen("qdsopts.txt", "r");
	
	int proper_argc;
	char **proper_argv;
	char command_line[256];
	memset(command_line, 0, 256);
	
	
	if (fp == NULL)
	{
		printf("Using default command line\n");
		proper_argc = 3;
		proper_argv = (char **)malloc(proper_argc * sizeof(char *));
		
		if (proper_argv == NULL)
		{
			printf("couldn't allocate space for argv\n");
			while(1);
		}
		
		int pos = 0;
		
		proper_argv[0] = command_line + pos;
		memcpy(command_line, "/_boot_mp.nds", strlen("/_boot_mp.nds") + 1);
		pos += strlen("/_boot_mp.nds") + 1;
		
		proper_argv[1] = command_line + pos;
		memcpy(command_line + pos, "-basedir", strlen("-basedir") + 1);
		pos += strlen("-basedir") + 1;
		
		proper_argv[2] = command_line + pos;
		memcpy(command_line + pos, "/", strlen("/") + 1);
		pos += strlen("/") + 1;
	}
	else
	{
		printf("Using command line from disk\n");
		fread(command_line + strlen("/_boot_mp.nds") + 1, 1, 256 - strlen("/_boot_mp.nds") - 1, fp);
		printf("command line is:\n%s\n", command_line);
		proper_argc = 0;
		
		for (int count = 0; count < 255; count++)		//yeah
		{
			//for windows etc
			if ((command_line[count] == 0x0d) && (command_line[count + 1] == 0x0a))
			{
				proper_argc++;
				count++;
			}
			//for unix
			else if (command_line[count + 1] == 0x0a)
			{
				proper_argc++;
			}
		}
		
		proper_argc++;		//for the program name
		proper_argv = (char **)malloc(proper_argc * sizeof(char *));
		int pos = 0;
		
		proper_argv[0] = command_line + pos;
		memcpy(command_line, "/_boot_mp.nds", strlen("/_boot_mp.nds") + 1);
		pos += strlen("/_boot_mp.nds") + 1;
	
		int arg_no = 1;
		
		for (int count = pos; count < 255; count++)		//yeah
		{
			//for windows etc
			if ((command_line[count] == 0x0d) && (command_line[count + 1] == 0x0a))
			{
				proper_argv[arg_no] = command_line + pos;
				pos += count - pos + 2;
				
				command_line[count] = 0;
				command_line[count + 1] = 0;
				
				arg_no++;
				count++;
			}
			//for unix
			else if (command_line[count] == 0x0a)
			{
				proper_argv[arg_no] = command_line + pos;
				pos += count - pos + 1;
				
				command_line[count] = 0;
				
				arg_no++;
			}
		}
		
		printf("Quake command line options:\n");
		for (int count = 1; count < proper_argc; count++)
			printf("%d:%s\n", count, proper_argv[count]);
	}

	frame_buffer_base = VRAM_A;
	frame_buffer_width = SCREEN_WIDTH;
	frame_buffer_height = SCREEN_HEIGHT;
	console_base = BG_GFX_SUB;
	
	ds_exram_unlock();
	ram_test();
	ds_exram_lock();
//	while(1);
	
	ds_exram_unlock();
	setGenericSound(11025, 127, 64, 1);
	init_arm7();
	
//	volatile int count = 0;
//	for (count = 0; count < 1000000; count++);
//	play_mp3(3);
//	
//	while(1);
	
//	cygprofile_enable();

	quake_main(proper_argc, proper_argv);
		
	return 0;
}

int count_largest_block(void)
{
	int largest = 0;
	void *ptr = NULL;
	
	do
	{
		if (ptr != NULL)
		{
			free(ptr);
			ptr = NULL;
			largest++;
		}
			
		ptr = malloc(largest + 1);
		
	} while(ptr != NULL);
	
	return largest;
}

int count_largest_block_kb(void)
{
	int largest = 0;
	void *ptr = NULL;
	
	do
	{
		if (ptr != NULL)
		{
			free(ptr);
			ptr = NULL;
			largest += 1024;
		}
			
		ptr = malloc(largest + 1024);
		
	} while(ptr != NULL);
	
	return largest;
}

void print_top_four_blocks_kb(void)
{
	void *l1, *l2, *l3, *l4;
	
	printf("big free blocks, kb:\n");
	int size = count_largest_block_kb();
	l1 = malloc(size);
	printf("1: %.2f\n", (float)size / 1024);
	
	size = count_largest_block_kb();
	l2 = malloc(size);
	printf("2: %.2f\n", (float)size / 1024);
	
	size = count_largest_block_kb();
	l3 = malloc(size);
	printf("3: %.2f\n", (float)size / 1024);
	
	size = count_largest_block_kb();
	l4 = malloc(size);
	printf("4: %.2f\n", (float)size / 1024);
	
	free(l1);
	free(l2);
	free(l3);
	free(l4);
	
	printf("done\n");
}

void print_top_four_blocks(void)
{
	void *l1, *l2, *l3, *l4;
	
	printf("big free blocks, kb:\n");
	int size = count_largest_block();
	l1 = malloc(size);
	printf("1: %.2f\n", (float)size / 1024);
	
	size = count_largest_block();
	l2 = malloc(size);
	printf("2: %.2f\n", (float)size / 1024);
	
	size = count_largest_block();
	l3 = malloc(size);
	printf("3: %.2f\n", (float)size / 1024);
	
	size = count_largest_block();
	l4 = malloc(size);
	printf("4: %.2f\n", (float)size / 1024);
	
	if (l1) free(l1);
	if (l2) free(l2);
	if (l3) free(l3);
	if (l4) free(l4);
	
	printf("done\n");
}

void get_pen_pos(short *px, short *py)
{
	touchPosition touchXY = touchReadXY();
	*px = touchXY.px;
	*py = touchXY.py;
}

int main(int argc, unsigned char **argv)
{
//	asm volatile ( 
//	  "mov sp, #0x02400000\n" 
//	  "sub sp, sp, #4096\n"
//	  "sub sp, sp, #4\n" 
//	);
	
	real_main(argc, argv);
	
	return 0;
}
